﻿using System;
using System.Collections.Generic;
using System.Linq;
using Clarion;
using Clarion.Framework;
using Clarion.Framework.Core;
using Clarion.Framework.Templates;
using WorldServerLibrary.Model;
using System.Threading;

namespace ClarionSimulation
{
    /// <summary>

    /// This event will be fired at the beginning of each cognitive cycle in order to get the current visual sensor information.
    /// </summary>
    /// <returns>The Visual Sensor Information Itens</returns> 
    public delegate IList<Thing> InputVisualSensorialInformationEventHandler();

    /// <summary>
    /// This event will be fired at the end of each cognitive cycle in order to translate the selected action to a World Server Command
    /// </summary>
    /// <param name="externalAction">The external action that was choice</param>
    public delegate void OutputActionChunkEventHandler(ClarionAgentActionType externalAction);

    /// <summary>
    /// Public enum that represents all possibilities of agent actions
    /// </summary>
    public enum ClarionAgentActionType
    {
        DO_NOTHING,
        ROTATE_CLOCKWISE,
        GO_AHEAD,
		MOVE_FOOD, 				//IA006
		MOVE_JEWEL,				//IA006
		EAT,					//IA006
		GET						//IA006
    }

    public class ClarionAgent
    {
        #region Constants

        /// <summary>
        /// Constant that represents the Visual Sensor
        /// </summary>
        private String SENSOR_VISUAL_DIMENSION = "VisualSensor";

        /// <summary>
        /// Constant that represents that there is at least one wall ahead
        /// </summary>
        private String DIMENSION_WALL_AHEAD = "WallAhead";

		/// IA006
		/// <summary>
		/// Joias no campo de visao
		/// </summary>
		private String DIMENSION_JEWEL_SEEING = "JewelSeeing";

		/// IA006
		/// <summary>
		/// Comida no campo de visao
		/// </summary>
		private String DIMENSION_FOOD_SEEING = "FoodSeeing";

		/// IA006
		/// <summary>
		/// Joia mais proxima
		/// </summary>
		private String DIMENSION_CLOSEST_JEWEL = "ClosestJewel";

		/// IA006
		/// <summary>
		/// Comida mais proxima
		/// </summary>
		private String DIMENSION_CLOSEST_FOOD = "ClosestFood";


        #endregion

        #region Properties

        #region Simulation

        /// <summary>
        /// If this value is greater than zero, the agent will have a finite number of cognitive cycle. Otherwise, it will have infinite cycles.
        /// </summary>
        public double MaxNumberOfCognitiveCycles = -1;

        /// <summary>
        /// Current cognitive cycle number
        /// </summary>
        private double CurrentCognitiveCycle = 0;

        /// <summary>
        /// Time between cognitive cycle in miliseconds
        /// </summary>
        public Int32 TimeBetweenCognitiveCycles = 0;

        /// <summary>
        /// A thread Class that will handle the simulation process
        /// </summary>
        private Thread runThread;

        #endregion

        #region Agent

        /// <summary>
        /// The agent 
        /// </summary>
        private Clarion.Framework.Agent CurrentAgent;

        /// <summary>
        /// For each cognitive cycle, this event will be called in order to the agent receives the current sensorial information
        /// </summary>
        public event InputVisualSensorialInformationEventHandler OnNewVisualSensorialInformation;

        /// <summary>
        /// For each cognitive cycle, this event will be called when the agent selects one action
        /// </summary>
        public event OutputActionChunkEventHandler OnNewExternalActionSelected;

        #endregion

        #region Perception Input

        /// <summary>
        /// Perception input to indicates a wall ahead
        /// </summary>
        private DimensionValuePair InputWallAhead;

		/// IA006
		/// <summary>
		/// Joia no sensor visual
		/// </summary>
		private DimensionValuePair InputJewelSeeing;

		/// IA006
		/// <summary>
		/// Joia mais proxima no sensor visual
		/// </summary>
		private DimensionValuePair InputClosestJewel;

		/// IA006
		/// <summary>
		/// Comida no sensor visual
		/// </summary>
		private DimensionValuePair InputFoodSeeing;

		/// IA006
		/// <summary>
		/// Comida mais proxima no sensor visual
		/// </summary>
		private DimensionValuePair InputClosestFood;


        #endregion

        #region Action Output

        /// <summary>
        /// Output action that makes the agent to rotate clockwise
        /// </summary>
        private ExternalActionChunk OutputRotateClockwise;

        /// <summary>
        /// Output action that makes the agent go ahead
        /// </summary>
        //private ExternalActionChunk OutputGoAhead;

		/// IA006
		/// <summary>
		/// Output para mover a criatura a uma joia
		/// </summary>
		private ExternalActionChunk OutputMoveJewel;

		/// IA006
		/// <summary>
		/// Output para pegar uma joia
		/// </summary>
		private ExternalActionChunk OutputGet;

		/// IA006
		/// <summary>
		/// Output para mover a criatura a uma coimda
		/// </summary>
		private ExternalActionChunk OutputMoveFood;

		/// IA006
		/// <summary>
		/// Output para comer
		/// </summary>
		private ExternalActionChunk OutputEat;



        #endregion

        #endregion

        #region Constructor

        public ClarionAgent()
        {
            // Initialize the agent
            CurrentAgent = World.NewAgent("Current Agent");

            // Initialize Input Information
            InputWallAhead = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_WALL_AHEAD);
			InputJewelSeeing = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_JEWEL_SEEING); //IA006
			InputClosestJewel = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_CLOSEST_JEWEL); //IA006
			InputFoodSeeing = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_FOOD_SEEING); //IA006
			InputClosestFood = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_CLOSEST_FOOD); //IA006

            // Initialize Output actions
            OutputRotateClockwise = World.NewExternalActionChunk(ClarionAgentActionType.ROTATE_CLOCKWISE.ToString());
            //OutputGoAhead = World.NewExternalActionChunk(ClarionAgentActionType.GO_AHEAD.ToString());
			OutputMoveJewel = World.NewExternalActionChunk (ClarionAgentActionType.MOVE_JEWEL.ToString()); //IA006
			OutputMoveFood = World.NewExternalActionChunk (ClarionAgentActionType.MOVE_FOOD.ToString()); //IA006
			OutputGet = World.NewExternalActionChunk (ClarionAgentActionType.GET.ToString()); //IA006
			OutputEat = World.NewExternalActionChunk (ClarionAgentActionType.EAT.ToString()); //IA006

            //Create thread to simulation
            runThread = new Thread(RunThread);
			Console.WriteLine("Agent started");
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Run the Simulation in World Server 3d Environment
        /// </summary>
        public void Run()
        {                
			Console.WriteLine ("Running ...");
            // Setup Agent to run
            if (runThread != null && !runThread.IsAlive)
            {
                SetupAgentInfraStructure();

                if (OnNewVisualSensorialInformation != null)
                {
                    // Start Simulation Thread                
                    runThread.Start(null);
                }
            }
        }

        /// <summary>
        /// Abort the current Simulation
        /// </summary>
        /// <param name="deleteAgent">If true beyond abort the current simulation it will die the agent.</param>
        public void Abort(Boolean deleteAgent)
        {   Console.WriteLine ("Aborting ...");
            if (runThread != null && runThread.IsAlive)
            {
                runThread.Abort();
            }

            if (CurrentAgent != null && deleteAgent)
            {
                CurrentAgent.Die();
            }
        }

        #endregion

        #region Setup Agent Methods

        /// <summary>
        /// Setup agent infra structure (ACS, NACS, MS and MCS)
        /// </summary>
        private void SetupAgentInfraStructure()
        {
            // Setup the ACS Subsystem
            SetupACS();                    
        }

        private void SetupMS()
        {            
            //RichDrive
        }

        /// <summary>
        /// Setup the ACS subsystem
        /// </summary>
        private void SetupACS()
        {
            // Create Colission Wall Rule
            SupportCalculator avoidCollisionWallSupportCalculator = FixedRuleDelegateToAvoidColissionWall;
            FixedRule ruleAvoidCollisionWall = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputRotateClockwise, avoidCollisionWallSupportCalculator);
            CurrentAgent.Commit(ruleAvoidCollisionWall);

            // Create Colission To Go Ahead
            //SupportCalculator goAheadSupportCalculator = FixedRuleDelegateToGoAhead;
            //FixedRule ruleGoAhead = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGoAhead, goAheadSupportCalculator);
            //CurrentAgent.Commit(ruleGoAhead);

			/// IA006 Rule Move Jewel
			SupportCalculator moveJewelSupportCalculator = FixedRuleDelegateToMoveJewel;
			FixedRule ruleMoveJewel = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputMoveJewel, moveJewelSupportCalculator);
			CurrentAgent.Commit(ruleMoveJewel);

			/// IA006 Rule Move Food
			SupportCalculator moveFoodSupportCalculator = FixedRuleDelegateToMoveFood;
			FixedRule ruleMoveFood = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputMoveFood, moveFoodSupportCalculator);
			CurrentAgent.Commit(ruleMoveFood);

			/// IA006 Rule Get
			SupportCalculator getSupportCalculator = FixedRuleDelegateToGet;
			FixedRule ruleGet = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGet, getSupportCalculator);
			CurrentAgent.Commit(ruleGet);

			/// IA006 Rule Eat
			SupportCalculator eatSupportCalculator = FixedRuleDelegateToEat;
			FixedRule ruleEat = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputEat, eatSupportCalculator);
			CurrentAgent.Commit(ruleEat);


            // Disable Rule Refinement
            CurrentAgent.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            // The selectio type will be probabilistic
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_METHOD = ActionCenteredSubsystem.LevelSelectionMethods.STOCHASTIC;

            // The selection of the action will be fixed (not variable) i.e. only the statement defined above.
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_OPTION = ActionCenteredSubsystem.LevelSelectionOptions.FIXED;

            // Define Probabilistic values
            CurrentAgent.ACS.Parameters.FIXED_FR_LEVEL_SELECTION_MEASURE = 1;
            CurrentAgent.ACS.Parameters.FIXED_IRL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_BL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_RER_LEVEL_SELECTION_MEASURE = 0;
        }

        /// <summary>
        /// Make the agent perception. In other words, translate the information that came from sensors to a new type that the agent can understand
        /// </summary>
        /// <param name="sensorialInformation">The information that came from server</param>
        /// <returns>The perceived information</returns>
        private SensoryInformation MakePerceptionFromSensorialInput(IList<Thing> sensorialInformation)
        {
            // New sensory information
            SensoryInformation si = World.NewSensoryInformation(CurrentAgent);

			//Console.WriteLine(sensorialInformation);
			Creature creature = (Creature) sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_CREATURE)).First();
			//creature.PrintLeaflet();	

            // Detect if we have a wall ahead
            //Boolean wallAhead = sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature <= 61)).Any();
            //double wallAheadActivationValue = wallAhead ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;
			//si.Add (InputWallAhead, wallAheadActivationValue);

			IEnumerable<Thing> sorted = sensorialInformation.Where(item => (item.CategoryId != Thing.CATEGORY_CREATURE && item.CategoryId != Thing.CATEGORY_BRICK));
			if (sorted.Any ()) {
				Thing closestObject = sorted.OrderBy (item => item.DistanceToCreature).First ();
				//Console.Out.WriteLine ("Closest Object found: " + closestObject.Name + "\n");

				// IA006 detecta a comida mais proxima "comivel"
				Boolean closestFood = sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_PFOOD && item.DistanceToCreature <= 30)).Any ();
				double closestFoodActivationValue = closestFood ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;

				// IA006 detecta a joia mais proxima "pegavel"
				Boolean closestJewel = sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_JEWEL && item.DistanceToCreature <= 30)).Any ();
				double closestJewelActivationValue = closestJewel && !closestFood ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;

				/// IA006 detecta comida
				Boolean foodSeeing = sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_PFOOD && item.DistanceToCreature > 30)).Any ();
				double foodSeeingActivationValue = foodSeeing && !closestFood && !closestJewel ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;
				
				/// IA006 detecta joias
				Boolean jewelSeeing = sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_JEWEL && creature.inLeaflet(item.Material.Color) && item.DistanceToCreature > 30)).Any ();
				double jewelSeeingActivationValue = jewelSeeing && !foodSeeing && !closestFood && !closestJewel ? CurrentAgent.Parameters.MAX_ACTIVATION : CurrentAgent.Parameters.MIN_ACTIVATION;

				
				si.Add (InputJewelSeeing, jewelSeeingActivationValue);
				si.Add (InputFoodSeeing, foodSeeingActivationValue);
				si.Add (InputClosestFood, closestFoodActivationValue);
				si.Add (InputClosestJewel, closestJewelActivationValue);
								
				if (!jewelSeeing && !foodSeeing && !closestFood && !closestJewel){
					si.Add (InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION);
				}else{
					si.Add (InputWallAhead, CurrentAgent.Parameters.MIN_ACTIVATION);
				}
				
			} else {
				si.Add (InputJewelSeeing, CurrentAgent.Parameters.MIN_ACTIVATION);
				si.Add (InputFoodSeeing, CurrentAgent.Parameters.MIN_ACTIVATION);
				si.Add (InputClosestFood, CurrentAgent.Parameters.MIN_ACTIVATION);
				si.Add (InputClosestJewel, CurrentAgent.Parameters.MIN_ACTIVATION);
				si.Add (InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION);
			}

			//IA006 Se si for vazio girar


            return si;
        }

        #endregion

        #region Delegate Methods

        #region Fixed Rules

        private double FixedRuleDelegateToAvoidColissionWall(ActivationCollection currentInput, Rule target)
        {
            // See partial match threshold to verify what are the rules available for action selection
            return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        }

        private double FixedRuleDelegateToGoAhead(ActivationCollection currentInput, Rule target)
        {
            // Here we will make the logic to go ahead
            return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MIN_ACTIVATION))) ? 1.0 : 0.0;
        }

		//IA006
		private double FixedRuleDelegateToMoveJewel(ActivationCollection currentInput, Rule target)
		{
			/// obter o elemento da visao e mover
			return ((currentInput.Contains(InputJewelSeeing, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}

		//IA006
		private double FixedRuleDelegateToMoveFood(ActivationCollection currentInput, Rule target)
		{
			/// obter o elemento da visao e mover
			return ((currentInput.Contains(InputFoodSeeing, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}

		//IA006
		private double FixedRuleDelegateToGet(ActivationCollection currentInput, Rule target)
		{
			/// pegar a joia
			return ((currentInput.Contains(InputClosestJewel, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}

		//IA006
		private double FixedRuleDelegateToEat(ActivationCollection currentInput, Rule target)
		{
			/// comer
			return ((currentInput.Contains(InputClosestFood, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}

        #endregion

        #region Run Thread Method

        private void RunThread(object obj)
        {
			Console.WriteLine("Starting Cognitive Cycle ... press CTRL-C to finish !");
            // Cognitive Cycle starts here getting sensorial information
            while (CurrentCognitiveCycle != MaxNumberOfCognitiveCycles)
            {   
                // Get current sensorial information                    
                IList<Thing> sensorialInformation = OnNewVisualSensorialInformation();

                // Make the perception
                SensoryInformation si = MakePerceptionFromSensorialInput(sensorialInformation);

                //Perceive the sensory information
                CurrentAgent.Perceive(si);

                //Choose an action
                ExternalActionChunk chosen = CurrentAgent.GetChosenExternalAction(si);

                // IA006 Get the selected action
                String actionLabel = chosen.LabelAsIComparable.ToString();
                ClarionAgentActionType actionType = (ClarionAgentActionType)Enum.Parse(typeof(ClarionAgentActionType), actionLabel, true);

                // Call the output event handler
                if (OnNewExternalActionSelected != null)
                {
                    OnNewExternalActionSelected(actionType);
                }

                // Increment the number of cognitive cycles
                CurrentCognitiveCycle++;

                //Wait to the agent accomplish his job
                if (TimeBetweenCognitiveCycles > 0)
                {
                    Thread.Sleep(TimeBetweenCognitiveCycles);
                }
            }
        }

        #endregion

        #endregion
    }
}
